# Datei: numpy-benning2.py
# Dieses Beispiel soll nur zeigen, wie einfach man mit NumPy und
# Matrizen arbeiten kann (ab Python-Variante 3.6). NumPy gehört
# nicht zu Standard-Python. Deshalb muss das Paket <numpy> separat
# u.U. erst in der Eingabeaufforderung (Konsolefenster) von
# Windows geladen werden durch:
#    python -m pip install numpy
#
# Als Beispiel dient hier die Helmerttransformation aus: 
# Benning, 4. Auflage, Statistik in Geodäsie, Geoinformation 
# und Bauwesen, Wichmann Verlag, 2011, S. 267 ff.
# in gleicher Syntax.
# Der zu benutzende Editor sollte mit UTF-8 lauffähig sein.

import numpy as np

print("")
print("--- Beispiel: Helmerttransformation mit NumPy")
print("")
print("--- Datenquelle: Benning, 4. Auflage, Statistik in Geodäsie,")
print("--- Geoinformation und Bauwesen, Wichmann Verlag, 2011, S. 259 ff., 267 ff.")
print("--- 4-Parameter-Helmerttransformation")
print("")
print("Dieses harmlose Beispiel soll nur verdeutlichen, wie einfach")
print("Matrizenoperationen in Python mit <NumPy> zu handhaben sind.")
print("Die lfd. Nummerierung soll dazu dienen, die Eingabe (= numpy-benning2.py)")
print("und diese Ausgabe direkt miteinander vergleichen zu können.")
print("Der zu benutzende Editor sollte mit UTF-8 lauffähig sein.")
print("")


def fett(text, trans=str.maketrans(
        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789",
        "𝗔𝗕𝗖𝗗𝗘𝗙𝗚𝗛𝗜𝗝𝗞𝗟𝗠𝗡𝗢𝗣𝗤𝗥𝗦𝗧𝗨𝗩𝗪𝗫𝗬𝗭𝗮𝗯𝗰𝗱𝗲𝗳𝗴𝗵𝗶𝗷𝗸𝗹𝗺𝗻𝗼𝗽𝗾𝗿𝘀𝘁𝘂𝘃𝘄𝘅𝘆𝘇𝟬𝟭𝟮𝟯𝟰𝟱𝟲𝟳𝟴𝟵")):
    return text.translate(trans)

p = 4
print("\n[01]\n  n (bzw. p = Anzahl der aktiven Passpunkte), hier n = 4")
print("  e (bzw. \u03B5 ) = -v (Residuum, Verbesserung)")
print("  "+fett("y")+" = "+fett("X")+"\u00b7\u03B2 + "+fett("e")+" (Beobachtungsgleichungen)")

yT= np.array([413.6, 377.6, 487.6, 706.4, 757.3, 456.2, 706.6, 686.2])
print("\n[02]\n  "+fett("y")+"\u1D40 (Koordinaten der Passpunkte \
im Zielsystem)\n", yT)

y=yT.T
print("\n[03]\n  "+fett("y")+" (Koordinaten der Passpunkte im Zielsystem)\n", y)

X = np.array([[1.,0.,-186.1850, -170.1375],
     [0.,1., 170.1375, -186.1850],
     [1.,0., 145.3350, -109.8275],
     [0.,1., 109.8275,  145.3350],
     [1.,0., -93.4750,  170.0625],
     [0.,1.,-170.0625,  -93.4750],
     [1.,0., 134.3250,  109.9025],
     [0.,1.,-109.9025,  134.3250]])
print("\n[04]\n  "+fett("X")+" (Funktional-, Koeffizienten-, Designmatrix mit\n\
 reduzierten Passpunktkoordinaten im Startsystem)\n", X)

N = np.transpose(X) @ X
print("\n[05]\n  "+fett("N")+" = "+fett("X")+"\u1D40"+fett("X")+\
      " (Normalgleichungsmatrix)\n", N)

print("\n[06]\n  "+fett("X")+"\u1D40"+fett("X")+"\u00b7\u03B2 = "+fett("N")+"\u00b7\u03B2 = "+fett("X")+\
      "\u1D40\u00b7"+fett("y")+" = "+fett("n")+" (Normalgleichungen)") 

Ninv = np.linalg.inv(N)
print("\n[07]\n  "+fett("N")+"\u207B\u00B9 (Inverse der \
Normalgleichungsmatrix)\n", Ninv)

n = np.transpose(X) @ y
print("\n[08]\n  "+fett("n")+" = "+fett("X")+"\u1D40\u00b7"+\
      fett("y")+" (rechte Seite der Normalgleichungen)\n", n)

beta= Ninv @ n
print("\n[09]\n  \u03B2 = "+fett("N")+"\u207B\u00B9\u00b7"+fett("n")+\
      " (Transformationsparameter)\n", beta)

masstab = np.hypot(beta[2],beta[3])
drehwinkel = np.arctan(beta[2]/beta[3]) * 200./np.pi
print("\n[10]\n Ma\u1E9Estabsfaktor = ", masstab, "\n Drehwinkel =", drehwinkel, "gon")

v= X @ beta - y
print("\n[11]\n  "+fett("v")+" = -"+fett("e")+" = "+fett("X")+\
      "\u03B2 - "+fett("y")+" (Verbesserugen, -Residuenvektor)\n", v)

ee = v @ v
print("\n[12]\n  "+fett("e")+"\u1D40"+fett("e")+\
      " (Residuenquadratsumme, Summe "+fett("v")+\
      "\u1D40"+fett("v")+")\n", ee)

ss = ee/(2*p - 4)
print("\n[13]\n  s\u00B2 = "+fett("e")+"\u1D40"+fett("e")+\
      "/(2n - 4) (Residuenvarianz der Gewichtseinheit)\n",ss)

s = np.sqrt(ss)
print("\n[14]\n  s\u2080 = \u221A("+fett("e")+"\u1D40"+fett("e")+\
      "/(2n - 4)) (empirische Standardabweichung der Gewichtseinheit)\n",s)

m = s*np.sqrt(2)
print ("\n  m\u2080 = s\u2080\u221A2 (mittlere restliche Klaffung eines Punktes)\n",m)


H = X @ Ninv @ np.transpose(X)
print("\n[15]\n  "+fett("H")+" = "+fett("XN")+"\u207B\u00B9"+fett("X")+\
      "\u1D40 (Projektionsmatrix, Hatmatrix)\n",H)

diagonale = np.diag(H)
print("\n[16]\n  Diagonale der Projektionsmatrix "+fett("H")+"\n",diagonale)

redundanzen = 1-diagonale
print("\n[17]")
print("  "+fett("F")+" = ("+fett("E")+" - "+fett("H")+\
      ") (Redundanzmatrix, "+fett("E")+" = Einheitsmatrix)")
print("  f\u1D62\u1D62 = r\u1D62 = 1 - h\u1D62\u1D62 (Redundanz)")
print("  Diagonale "+fett("F")+" (Redundanzen)\n",redundanzen)

print("\n",60*'-')

sigma_beta =  Ninv*ss
print("\n[18]\n  \u03A3\u03B2 = s\u00B2\u00b7"+fett("N")+\
      "\u207B\u00B9\n (Kovarianzmatrix, Varianz-Kovarianzmatrix oder\
 Dispersionsmatrix\n  der Parameterschätzwerte)\n",sigma_beta)

diag_sigma = np.diag(sigma_beta)
print("\n[19]\n  \u03A3\u03B2 = s\u00B2\u00b7"+fett("N")+\
      "\u207B\u00B9\n \
(Varianzen der Mittelwerte und Parameterschätzwerte)\n",diag_sigma)

sqrt_diag_sigma = np.sqrt(diag_sigma)
print("\n[20]\n  [sY, sX, s\u2092, s\u2090] = \u221A(\u03A3\u03B2)\n \
(Standardabweichungen der Mittelwerte und \
Parameterschätzwerte\n",sqrt_diag_sigma)

print("\n",60*'-')

sigma_y = H*ss
print("\n[21]\n  \u03A3y = s\u00B2\u00b7" + fett("XN")+"\u207B\u00B9"+fett("X")+\
      "\u1D40  = s\u00B2\u00b7"+fett("H")+"\n (Kovarianzmatrix der \
Schätzwerte der Passpunktkoordinaten Y\u1D62, X\u1D62)\n",sigma_y)

diag_sigma_y = np.diag(sigma_y)
print("\n[22]\n  Varianzen der Schätzwerte Y\u1D62, X\u1D62\n",diag_sigma_y)

#Berechnung der Standardabweichungen
sqrt_diag_sigma_y = np.sqrt(diag_sigma_y)
print("\n[23]\n  1. Variante: Standardabweichungen der Schätzwerte der\n \
Passpunktkoordinaten Y\u1D62, X\u1D62\n",sqrt_diag_sigma_y)

std_deviations = np.sqrt(np.diag(sigma_y))
print("\n[24]\n  2. Variante: Standardabweichungen der Schätzwerte der\n \
Passpunktkoordinaten Y\u1D62, X\u1D62\n",std_deviations)

print("\n",60*'-')

#Berechnen der Korrelationskoeffizienten
korrelation_sigma_y = sigma_y / (std_deviations[:, np.newaxis]*std_deviations)
print("\n[25]\n  "+fett("R")+" (Korrelationsmatrix, aus s\u00B2\u00b7"+fett("H")+\
      ")\n 1. Variante: Korrelationsmatrix der Schätzwerte der\n \
Passpunktkoordinaten Y\u1D62, X\u1D62\n",korrelation_sigma_y)

print ("\n  Vorbereitungen zur 2. Variante:")

#Erstellen einer leeren Korrelationsmatrix
korrelation2_sigma_y = np.zeros_like(sigma_y)
print("\n[26]\n  leere Korrelationsmatrix\n", korrelation2_sigma_y)

for i in range(len(sigma_y)):
    for j in range(len(sigma_y)):
        korrelation2_sigma_y[i, j] = sigma_y[i, j]\
                      / (std_deviations[i] * std_deviations[j])
print("\n[27]\n  "+fett("R")+" (aus s\u00B2\u00b7"+fett("H")+\
")\n 2. Variante: Korrelationsmatrix der Schätzwerte der\n \
Passpunktkoordinaten Y\u1D62, X\u1D62\n", korrelation2_sigma_y)

print ("\n  Vorbereitungen zur 3. Variante:")

matrixdiag3 = np.diag(diag_sigma_y)
print ("\n[28]\n  Nullmatrix mit voll besetzter Diagonale \
(Varianzen)\n",matrixdiag3)
       
matrixdiag = np.diag(std_deviations)
print ("\n[29]\n  Nullmatrix mit voll besetzter Diagonale \
(Standardabweichungen)\n",matrixdiag)

matrixdiag2 =  np.linalg.inv(matrixdiag)
print ("\n[30]\n  "+fett("S")+" (Nullmatrix mit voll besetzter, inverser \
Diagonale)\n (reziproke Standardabweichungen)\n",matrixdiag2)

korrelation_sigma_y3 = matrixdiag2 @sigma_y @matrixdiag2
print("\n[31]\n  "+fett("R")+" = "+fett("S")+"\u00b7\u03A3y\u00b7"+fett("S")+\
      "\n 3.Variante: Korrelationsmatrix der Schätzwerte der\n \
Passpunktkoordinaten Y\u1D62, X\u1D62\n",korrelation_sigma_y3)

print("\n[32]\n  Die Funktion np.corrcoef() zur Erzeugung der \
Korrelationsmatrix\n der Schätzwerte ergibt keine identische Ausgabe.")

print("\n",60*'-')

Neu = np.array([[1.,0.,-16.9050, -10.1675],
     [0.,1., 10.1675, -16.9050],[1.,0., 0., 0.],
    [0.,1., 0., 0.]])
print("\n[33]\n  "+fett("Neu")+"\n (zu transformierende Koordinaten \
im Startsystem, Schwerpunkt)\n", Neu)

TNeu = Neu @ beta
print("\n[34]\n  ["+fett("Y")+"\u2C7C, "+fett("X")+"\u2C7C] = " +\
      fett("Neu")+"\u00b7\u03B2 (Transformationsgleichungen)\n  (Neupunktkoordinaten j im \
Zielsystem, Schwerpunkt)\n", TNeu)

#Erstellen eines leeren Vektors für die Varianzen der
#Neupunktkoordinaten
VNeu = np.zeros_like(TNeu)
#Anzahl der Zeilen
shape=VNeu.shape[0]
element1=Ninv[p-1][p-1]
for i in range (shape):
    element2=Neu[i][2]
    element3=Neu[i][3]
    VNeu[i]=ss*(1 +1/p + ((element2)**2 + (element3)**2)*element1)    
    
print("\n[35]\n  Varianzen der Neupunktkoordinaten j\n", VNeu)

print("\n[36]\n  Standardabweichungen der Neupunktkoordinaten j\n \
(ohne Berücksichtigung einer eventuellen \
Restklaffenverteilung)\n", np.sqrt(VNeu))
      
print("\n",60*'=')

